@@ -32,7 +32,8 @@
  */
 
 #include "ldb_tdb.h"
-#include <lib/tdb_compat/tdb_compat.h>
+#include "ldb_private.h"
+#include <tdb.h>
 
 /*
   add one element to a message
@@ -212,7 +213,8 @@ static int ltdb_search_base(struct ldb_module *module, struct ldb_dn *dn)
 {
 	void *data = ldb_module_get_private(module);
 	struct ltdb_private *ltdb = talloc_get_type(data, struct ltdb_private);
-	TDB_DATA tdb_key, tdb_data;
+	TDB_DATA tdb_key;
+	int exists;
 
 	if (ldb_dn_is_null(dn)) {
 		return LDB_ERR_NO_SUCH_OBJECT;
@@ -224,14 +226,33 @@ static int ltdb_search_base(struct ldb_module *module, struct ldb_dn *dn)
 		return LDB_ERR_OPERATIONS_ERROR;
 	}
 
-	tdb_data = tdb_fetch_compat(ltdb->tdb, tdb_key);
+	exists = tdb_exists(ltdb->tdb, tdb_key);
 	talloc_free(tdb_key.dptr);
-	if (!tdb_data.dptr) {
-		return LDB_ERR_NO_SUCH_OBJECT;
+		
+	if (exists) {
+		return LDB_SUCCESS;
 	}
-	
-	free(tdb_data.dptr);
-	return LDB_SUCCESS;
+	return LDB_ERR_NO_SUCH_OBJECT;
+}
+
+struct ltdb_parse_data_unpack_ctx {
+	struct ldb_message *msg;
+	struct ldb_module *module;
+};
+
+static int ltdb_parse_data_unpack(TDB_DATA key, TDB_DATA data,
+				  void *private_data)
+{
+	struct ltdb_parse_data_unpack_ctx *ctx = private_data;
+
+	struct ldb_context *ldb = ldb_module_get_ctx(ctx->module);
+	int ret = ldb_unpack_data(ldb, (struct ldb_val *)&data, ctx->msg);
+	if (ret == -1) {
+		ldb_debug(ldb, LDB_DEBUG_ERROR, "Invalid data for index %*.*s\n",
+			  (int)key.dsize, (int)key.dsize, key.dptr);
+		return LDB_ERR_OPERATIONS_ERROR;		
+	}
+	return ret;
 }
 
 /*
@@ -246,9 +267,11 @@ int ltdb_search_dn1(struct ldb_module *module, struct ldb_dn *dn, struct ldb_mes
 	void *data = ldb_module_get_private(module);
 	struct ltdb_private *ltdb = talloc_get_type(data, struct ltdb_private);
 	int ret;
-	TDB_DATA tdb_key, tdb_data;
-
-	memset(msg, 0, sizeof(*msg));
+	TDB_DATA tdb_key;
+	struct ltdb_parse_data_unpack_ctx ctx = {
+		.msg = msg,
+		.module = module
+	};
 
 	/* form the key */
 	tdb_key = ltdb_key(module, dn);
@@ -256,24 +279,24 @@ int ltdb_search_dn1(struct ldb_module *module, struct ldb_dn *dn, struct ldb_mes
 		return LDB_ERR_OPERATIONS_ERROR;
 	}
 
-	tdb_data = tdb_fetch_compat(ltdb->tdb, tdb_key);
-	talloc_free(tdb_key.dptr);
-	if (!tdb_data.dptr) {
-		return LDB_ERR_NO_SUCH_OBJECT;
-	}
-	
+	memset(msg, 0, sizeof(*msg));
+
 	msg->num_elements = 0;
 	msg->elements = NULL;
 
-	ret = ltdb_unpack_data(module, &tdb_data, msg);
-	free(tdb_data.dptr);
+	ret = tdb_parse_record(ltdb->tdb, tdb_key, 
+			       ltdb_parse_data_unpack, &ctx); 
+	talloc_free(tdb_key.dptr);
+	
 	if (ret == -1) {
-		struct ldb_context *ldb = ldb_module_get_ctx(module);
-		ldb_debug(ldb, LDB_DEBUG_ERROR, "Invalid data for index %s\n",
-			  ldb_dn_get_linearized(msg->dn));
-		return LDB_ERR_OPERATIONS_ERROR;		
+		if (tdb_error(ltdb->tdb) == TDB_ERR_NOEXIST) {
+			return LDB_ERR_NO_SUCH_OBJECT;
+		}
+		return LDB_ERR_OPERATIONS_ERROR;
+	} else if (ret != LDB_SUCCESS) {
+		return ret;
 	}
-
+	
 	if (!msg->dn) {
 		msg->dn = ldb_dn_copy(msg, dn);
 	}
@@ -384,10 +407,18 @@ int ltdb_filter_attrs(struct ldb_message *msg, const char * const *attrs)
 	}
 
 	talloc_free(msg->elements);
-	msg->elements = talloc_realloc(msg, el2, struct ldb_message_element, msg->num_elements);
+
+	if (num_elements > 0) {
+		msg->elements = talloc_realloc(msg, el2, struct ldb_message_element,
+					       num_elements);
+	} else {
+		msg->elements = talloc_array(msg, struct ldb_message_element, 0);
+		talloc_free(el2);
+	}
 	if (msg->elements == NULL) {
 		return -1;
 	}
+
 	msg->num_elements = num_elements;
 
 	return 0;
@@ -414,13 +445,15 @@ static int search_func(struct tdb_context *tdb, TDB_DATA key, TDB_DATA data, voi
 
 	msg = ldb_msg_new(ac);
 	if (!msg) {
+		ac->error = LDB_ERR_OPERATIONS_ERROR;
 		return -1;
 	}
 
 	/* unpack the record */
-	ret = ltdb_unpack_data(ac->module, &data, msg);
+	ret = ldb_unpack_data(ldb, (struct ldb_val *)&data, msg);
 	if (ret == -1) {
 		talloc_free(msg);
+		ac->error = LDB_ERR_OPERATIONS_ERROR;
 		return -1;
 	}
 
@@ -429,6 +462,7 @@ static int search_func(struct tdb_context *tdb, TDB_DATA key, TDB_DATA data, voi
 				     (char *)key.dptr + 3);
 		if (msg->dn == NULL) {
 			talloc_free(msg);
+			ac->error = LDB_ERR_OPERATIONS_ERROR;
 			return -1;
 		}
 	}
@@ -438,6 +472,7 @@ static int search_func(struct tdb_context *tdb, TDB_DATA key, TDB_DATA data, voi
 				  ac->tree, ac->base, ac->scope, &matched);
 	if (ret != LDB_SUCCESS) {
 		talloc_free(msg);
+		ac->error = LDB_ERR_OPERATIONS_ERROR;
 		return -1;
 	}
 	if (!matched) {
@@ -450,6 +485,7 @@ static int search_func(struct tdb_context *tdb, TDB_DATA key, TDB_DATA data, voi
 
 	if (ret == -1) {
 		talloc_free(msg);
+		ac->error = LDB_ERR_OPERATIONS_ERROR;
 		return -1;
 	}
 
@@ -457,6 +493,7 @@ static int search_func(struct tdb_context *tdb, TDB_DATA key, TDB_DATA data, voi
 	if (ret != LDB_SUCCESS) {
 		ac->request_terminated = true;
 		/* the callback failed, abort the operation */
+		ac->error = LDB_ERR_OPERATIONS_ERROR;
 		return -1;
 	}
 
@@ -474,6 +511,7 @@ static int ltdb_search_full(struct ltdb_context *ctx)
 	struct ltdb_private *ltdb = talloc_get_type(data, struct ltdb_private);
 	int ret;
 
+	ctx->error = LDB_SUCCESS;
 	if (ltdb->in_transaction != 0) {
 		ret = tdb_traverse(ltdb->tdb, search_func, ctx);
 	} else {
@@ -484,7 +522,7 @@ static int ltdb_search_full(struct ltdb_context *ctx)
 		return LDB_ERR_OPERATIONS_ERROR;
 	}
 
-	return LDB_SUCCESS;
+	return ctx->error;
 }
 
 /*
@@ -583,7 +621,7 @@ int ltdb_search(struct ltdb_context *ctx)
 				/* useful for debugging when slow performance
 				 * is caused by unindexed searches */
 				char *expression = ldb_filter_from_tree(ctx, ctx->tree);
-				ldb_debug(ldb, LDB_DEBUG_WARNING, "ldb FULL SEARCH: %s SCOPE: %s DN: %s\n",
+				ldb_debug(ldb, LDB_DEBUG_ERROR, "ldb FULL SEARCH: %s SCOPE: %s DN: %s",
 							expression,
 							req->op.search.scope==LDB_SCOPE_BASE?"base":
 							req->op.search.scope==LDB_SCOPE_ONELEVEL?"one":
